#include <math.h>
#include <iostream>
#include <vector>
#include <map>
#include "globals.h"
#include "revHash48.h"
#include "hash.h"

using namespace std;

#define MAX_KEY       255

RevHashTable48::RevHashTable48(int seed, int inxBits)
{
  if( (inxBits % NUM_DIVISIONS48) != 0 ) {
    cerr << "Number of index bits must be divisible by "
	 << NUM_DIVISIONS48 << endl;
    exit(1);
  }

  numBins = (1U << inxBits);
  bitsPerDiv = inxBits / NUM_DIVISIONS48;
  binsPerDiv = (1U << bitsPerDiv);

  // Initialize value table to all 0's
  valueTable.resize(numBins, 0);
    
  // Initialize lookup tables
  lookupTable.resize(NUM_DIVISIONS48);

  for( int i = 0; i < (int)NUM_DIVISIONS48; i++ ) {
    revLookupTable[i].resize(binsPerDiv);
    PopulateLookupTable(i, (seed+1) * (i+1));
  }
}

void RevHashTable48::AddValue(KeyType48 key, ValueType val)
{
  valueTable[GetIndex(key)] += val;
}

void RevHashTable48::RemoveValue(KeyType48 key, ValueType val)
{
  valueTable[GetIndex(key)] -= val;
}

ValueType RevHashTable48::GetValue(KeyType48 key)
{
  return valueTable[GetIndex(key)];
}

IndexType RevHashTable48::GetIndex(KeyType48 key)
{
  int index = 0;
  
  unsigned char *keyDivided = (unsigned char *)&key;
  for( int i = 0; i < (int)NUM_DIVISIONS48; i++ ) {
    int bin = (int)lookupTable[i][keyDivided[i]];
    index += bin << (i * bitsPerDiv);
  }

  return index;
}

void RevHashTable48::PopulateLookupTable(int div, int seed)
{
  // Initialize the random number generator
  srand(seed + 1);
  
  unsigned char bin;
  for( int i = 0; i < MAX_KEY + 1; i++ ) {
    bin = (int)((double)binsPerDiv*rand()/(RAND_MAX+1.0));
    
    lookupTable[div].push_back(bin);
    revLookupTable[div][bin].push_back(i);
  }
}

void RevHashTable48::ChangeDetection(list<IndexType> &changes,
				   const RevHashTable48 &otherTbl, 
				   double threshold)
{
  changes.clear();
  
  ValueType difference;
  for( unsigned int i = 0; i < numBins; i++ ) {
    difference = valueTable[i] - otherTbl.valueTable[i];
    
    if( difference > threshold ) {
      changes.push_back(i);
    }
  }
}

double RevHashTable48::FindThreshold(const RevHashTable48 &otherTbl,
				   int numAnoms, double minThreshold)
{
  ValueType difference;

  if( numAnoms == -1 )
    return minThreshold;

  // Create a map with a key of the change mag. and a value of the corr. bin
  list<ValueType> chgLst;

  for( unsigned int i = 0; i < numBins; i++ ) {
    difference = valueTable[i] - otherTbl.valueTable[i];
    
    if( difference > minThreshold ) {
      chgLst.push_back(difference);
    }
  }

  if( chgLst.size() == 0 )
    return -1.0;
  else if( chgLst.size() <= (unsigned int)numAnoms )
    return minThreshold;
  else {
    chgLst.sort();  // Sort the list of changes

    list<ValueType>::reverse_iterator rIt = chgLst.rbegin();
    for( int anom = 0; anom < numAnoms; anom++ )
      rIt++;

    return *rIt - 1.0;
  }
}

void RevHashTable48::GetPossibleKeys(list<unsigned char> &keys,
				   int div, IndexType bin)
{
  int mask = 0;
  for( unsigned int bit = 0; bit < bitsPerDiv; bit++ ) {
    mask <<= 1;
    mask |= 0x1;
  }

  bin = (bin >> (div * bitsPerDiv)) & mask;

  int numKeys = revLookupTable[div][bin].size();
  keys.clear();

  for( int i = 0; i < numKeys; i++ )
    keys.push_back(revLookupTable[div][bin][i]);
}

