#include <math.h>
#include <map>
#include <set>
#include "revKSketch48.h"
//#include "revKSketch.h"

#include "mangler.h"
#include "globals.h"

using namespace std;


RevKSketch48::RevKSketch48(int numHashTbls, int myR, int inxBits, int mySize)
{
  numTbls = numHashTbls;
  r = myR;
  iterSize = mySize;

  extraSketch = new KSketch(numHashTbls, inxBits);

  // Populate data tables
  for( int i = 0; i < numHashTbls; i++ )
    hashTbls.push_back(RevHashTable48(i, inxBits));

  sum = 0;  

#ifdef NEW_MANGLE
  mangler = new Mangler();
#else
  a1 = A1;
  a2 = A2;

  a1_inv = FindInverse(a1);
  a2_inv = FindInverse(a2);
#endif
}

RevKSketch48::~RevKSketch48()
{
#ifdef NEW_MANGLE
  delete mangler;
#endif
  delete extraSketch;
}

void RevKSketch48::AddValue(KeyType48 key, ValueType val)
{
  KeyType48 mangled = ReorderKey(key);

  for( int i = 0; i < numTbls; i++ )
    hashTbls[i].AddValue(mangled, val);

  extraSketch->AddValue(key, val);

  sum += val;
}

void RevKSketch48::RemoveValue(KeyType48 key, ValueType val)
{
  KeyType48 mangled = ReorderKey(key);

  for( int i = 0; i < numTbls; i++ )
    hashTbls[i].RemoveValue(mangled, val);

  //extraSketch->RemoveValue(key, val);

  sum -= val;
}

ValueType RevKSketch48::EstValue(KeyType48 key)
{
  // Create a vector containing the value from each of the hash tables
  vector<ValueType> values;

  int k = hashTbls[0].GetNumBins();
  for( int i = 0; i < numTbls; i++ ) {
    ValueType est = hashTbls[i].GetValue(ReorderKey(key));

    est -= sum / k;
    est /= 1 - 1/k;
    
    values.push_back(est);
  }

  // Find and return the median of the resulting vector
  sort(values.begin(), values.end());

  if( (numTbls % 2) == 0 )
    // number of entries is even, average the two middle entries
    return (values[numTbls/2] + values[(numTbls-1)/2]) / 2;
  else
    // number of entries is odd, return the middle entry
    return values[numTbls/2];
}

void RevKSketch48::ChangeDetection(list<KeyType48> &resultList,
				 RevKSketch48 &otherSketch, double threshold)
{
  bool done = false;

  cerr << "Starting change detection..." << endl;

  // Get the heavy change buckets for each table
  vector<list<IndexType> > bktLists(numTbls);
  int iter = 0;
  
  while( !done && (iter < MAX_ITER) ) {
    cerr << "\tStarting iteration..." << endl;
    iter++;

    int tbl;
    ValueType nextThresh;
    ValueType maxThresh = 0;


    // Find a threshold (maxThresh) such that no table will return
    // more than ITER_ANOM anomalies
    for( tbl = 0; tbl < numTbls; tbl++ ) {
      bktLists[tbl].clear();

      nextThresh = hashTbls[tbl].FindThreshold(otherSketch.hashTbls[tbl],
					       iterSize, threshold);
      
      if( nextThresh == -1.0 )
	done = true;
      else if( nextThresh > maxThresh )
	maxThresh = nextThresh;
    }
    
    // Get the list of buckets corresponding to the threshold determined
    // above.
    for( tbl = 0; tbl < numTbls && !done; tbl++ ) {
      cerr << "tbl " << tbl << endl;
      hashTbls[tbl].ChangeDetection(bktLists[tbl], 
				    otherSketch.hashTbls[tbl],
				    maxThresh);
    }
    
    // Perform an iteration of the key solution algorithm
    if( !done ) {
      // Create the nodes in the graph
      GraphType graph;
      graph.clear();
      CreateNodes(graph, bktLists);
      
      // Create the links between each level
      for( uint32 div = 0; div < NUM_DIVISIONS48 - 1; div++ ) {
	CreateLinks(graph, div);
      }
    
      // Get the full keys from the graph after the traversal
      list<KeyType48> tempLst;
      GenerateFullKeys(graph, tempLst);

      done = true;
      list<KeyType48>::iterator keyIt;

      // Test each reported keys in the extra sketch
      for( keyIt = tempLst.begin(); keyIt != tempLst.end(); keyIt++ ) {
	ValueType extraEst = extraSketch->EstValue(*keyIt);
	ValueType othExtraEst = otherSketch.extraSketch->EstValue(*keyIt);
	  
	if( (extraEst - othExtraEst) > threshold ) {
	  resultList.push_back(*keyIt);
	  RemoveValue(*keyIt, extraEst);
	  otherSketch.RemoveValue(*keyIt, othExtraEst);
	  done = false;  // Found another key, so keep going
	}
      }
    }
  }

  resultList.sort();
  resultList.unique();
}

void RevKSketch48::CreateNodes(GraphType &graph,
			     const vector<list<IndexType> > &bktLists)
{
  // A map that associates key segments with nodes
  map<unsigned char, NodeType*> nodeMap;

  graph.resize(NUM_DIVISIONS48);  // Initialize the graph

  // For each key division
  for( uint32 div = 0; div < NUM_DIVISIONS48; div++ ) {
    nodeMap.clear();

    // For each hash table in the sketch
    for( int tbl = 0; tbl < numTbls; tbl++ ) {
    
      // For each heavy change bucket for this hash table
      for( list<IndexType>::const_iterator bktListIt = bktLists[tbl].begin(); 
	   bktListIt != bktLists[tbl].end(); 
	   bktListIt++ ) {

	list<unsigned char> keys;
	hashTbls[tbl].GetPossibleKeys(keys, div, *bktListIt);
	
	// For each of the possible key segments
	for( list<unsigned char>::iterator keyIt = keys.begin();
	     keyIt != keys.end();
	     keyIt++ ) {
	  
	  NodeType *node = nodeMap[*keyIt];  // Get the node, if one exists
	  
	  if( node == NULL ) { // Key does not already exist in the map
	    // Create a new node and add it to the map
	    node = new NodeType(*keyIt, *bktListIt, tbl, numTbls);
	    nodeMap[*keyIt] = node;
	  }
	  else {
	    if( node->bktLists[tbl].empty() ) // First entry for this HT
	      node->numFound++;

	    node->bktLists[tbl].push_back(*bktListIt);
	  }
	}  // End for ... keyIt
      }  // End for ... bktListIt
    }  // End for ... tbl

    // Enter the nodes where numTbls - numFound <= R into the graph
    for( map<unsigned char, NodeType*>::iterator nodeMapIt = nodeMap.begin();
	 nodeMapIt != nodeMap.end();
	 nodeMapIt++ ) {
      NodeType *node = nodeMapIt->second;

      if( (numTbls - node->numFound) <= r ) {
	graph[div].push_back(node);
      }
      else
	delete node;
    }  // End for ... nodeMapIt
    cerr << "graph nodes is div " << div << " = " << graph[div].size() << endl;

  }  // End for ... div
}

void RevKSketch48::CreateLinks(GraphType &graph, int startDiv)
{
  if( startDiv >= ((int)(NUM_DIVISIONS48) - 1) ) 
    return;  // cannot start at the last level

  list<NodeType*> *l1List = &(graph[startDiv]);
  list<NodeType*> *l2List = &(graph[startDiv+1]);
  
  list<NodeType*> linkList;

  for( list<NodeType*>::iterator l1It = l1List->begin();
       l1It != l1List->end();
       l1It++ ) {
    
    for( list<NodeType*>::iterator l2It = l2List->begin();
	 l2It != l2List->end();
	 l2It++ ) {

      NodeType *newNode = RIntersect(*l1It, *l2It);
      if( newNode != NULL )
	linkList.push_back(newNode);
    }  // End for...l2It

    delete *l1It;  // no longer need this node in level 1
  }

  // Delete all of the nodes in L2
  for( list<NodeType*>::iterator l2It = l2List->begin();
       l2It != l2List->end();
       l2It++ )
    delete *l2It;

  l1List->clear();
  l2List->clear();
  l2List->insert(l2List->end(), linkList.begin(), linkList.end());
}

NodeType* RevKSketch48::RIntersect(NodeType *node1, NodeType *node2)
{
  NodeType *newNode = new NodeType;
  newNode->bktLists.resize(numTbls);

  for( int tbl = 0; tbl < numTbls; tbl++ ) {
    list<IndexType> *node1Lst = &(node1->bktLists[tbl]);
    list<IndexType> *node2Lst = &(node2->bktLists[tbl]);
    list<IndexType> *newLst = &(newNode->bktLists[tbl]);

    ListIntersect(*newLst, *node1Lst, *node2Lst);

    if( newLst->empty() ) {
      if( newNode->numFound >= r ) {
	delete newNode;
	return NULL;
      }
      else
	newNode->numFound++;
    }
  }
  
  list<unsigned char> *keyLst = &(newNode->keyList);
  keyLst->insert(keyLst->end(), node1->keyList.begin(), node1->keyList.end());
  keyLst->insert(keyLst->end(), node2->keyList.begin(), node2->keyList.end());
    
  return newNode;
}

void RevKSketch48::GenerateFullKeys(GraphType &graph, list<KeyType48> &resultList)
{
  KeyType48 fullKey;
  unsigned char *keyDivided = (unsigned char *)&fullKey;

  for( list<NodeType*>::iterator lstIt = graph[NUM_DIVISIONS48 - 1].begin();
       lstIt != graph[NUM_DIVISIONS48 - 1].end();
       lstIt++ ) {

    list<unsigned char> *keyList = &((*lstIt)->keyList);
    if( keyList->size() == NUM_DIVISIONS48 ) {
      int div = 0;
      for( list<unsigned char>::iterator keyIt = keyList->begin();
	   keyIt != keyList->end();
	   keyIt++, div++ ) {
	
	keyDivided[div] = *keyIt;
      }
      
      resultList.push_back(RestoreKey(fullKey));
    }  // end if

    delete *lstIt;
  }  // end for
}

void RevKSketch48::ListIntersect(list<IndexType> &outList,
		   const list<IndexType> &l1,
		   const list<IndexType> &l2)
{
  outList.clear();

  list<IndexType>::const_iterator l1It = l1.begin();
  list<IndexType>::const_iterator l2It = l2.begin();

  while( true ) {
    if( (l1It == l1.end()) || (l2It == l2.end()) )
      return;
    
    if( *l1It > *l2It )
      l2It++;
    else if( *l1It < *l2It )
      l1It++;
    else {
      outList.push_back(*l1It);
      l1It++;
      l2It++;
    }
  }
}

KeyType48 RevKSketch48::ReorderKey(KeyType48 key)
{
#ifdef NEW_MANGLE
  return mangler->MangleCharTable(key);
#else
  KeyType48 mangled = 0;
  unsigned int *keyDivided = (unsigned int *)&key;

  mangled  = (KeyType48)((keyDivided[1] * a1) % N) << 32;
  mangled += (KeyType48)((keyDivided[0] * a2) % N);
  
  return mangled;
#endif
}

KeyType48 RevKSketch48::RestoreKey(KeyType48 key)
{
#ifdef NEW_MANGLE
  return mangler->ReverseCharTable(key);
#else
  KeyType48 restored = 0;
  unsigned int *keyDivided = (unsigned int *)&key;

  restored  = (KeyType48)((keyDivided[1] * a1_inv) % N) << 32;
  restored += (KeyType48)((keyDivided[0] * a2_inv) % N);

  return restored;
#endif
}

#ifndef NEW_MANGLE
typedef struct ExtEuclidType {
  long long d;
  long long x;
  long long y;

  ExtEuclidType(long long myD, long long myX, long long myY)
    : d(myD), x(myX), y(myY) {}

  ExtEuclidType() {}
} ExtEuclidType;
  
ExtEuclidType ExtendedEuclid(long long a, long long b)
{
  if( b == 0 )
    return ExtEuclidType(a, 1, 0);
  
  ExtEuclidType extEuclidPrime = ExtendedEuclid(b, a % b);
  
  ExtEuclidType extEuclid;
  extEuclid.d = extEuclidPrime.d;
  extEuclid.x = extEuclidPrime.y;
  extEuclid.y = extEuclidPrime.x - (a / b)*extEuclidPrime.y;

  return extEuclid;
}
  
unsigned int RevKSketch48::FindInverse(unsigned int a)
{
  return (unsigned int)((ExtendedEuclid(a, N)).x % N);
}
#endif

IndexType RevKSketch48::GetIndex(KeyType48 key, int tbl)
{
  return hashTbls[tbl].GetIndex(ReorderKey(key));
}
